from dataclasses import dataclass, field
from typing import Any, Awaitable, Callable, TypeAlias, TypeVar

from latch_o11y.o11y import (
    AttributesDict,
    app_tracer,
    dict_to_attrs,
    trace_app_function,
)

from .asgi_iface import HTTPReceiveCallable, HTTPScope, HTTPSendCallable
from .auth import Authorization, get_signer_sub
from .framework import HTTPMethod, current_http_request_span, receive_class_ext

T = TypeVar("T")


@dataclass
class Context:
    scope: HTTPScope
    receive: HTTPReceiveCallable
    send: HTTPSendCallable

    auth: Authorization = field(default_factory=Authorization, init=False)

    _header_cache: dict[bytes, bytes] = field(default_factory=dict, init=False)
    _db_response_idx: int = field(default=0, init=False)

    def __post_init__(self):
        with app_tracer.start_as_current_span("find Authentication header"):
            auth_header = self.header_str("authorization")

        if auth_header is not None:
            self.auth = get_signer_sub(auth_header)

        if self.auth.oauth_sub is not None:
            current_http_request_span().set_attribute("enduser.id", self.auth.oauth_sub)

    def header(self, x: str | bytes):
        if isinstance(x, str):
            x = x.encode("utf-8")

        if x in self._header_cache:
            return self._header_cache[x]

        for k, v in self.scope.headers:
            self._header_cache[k] = v
            if k == x:
                return v

        return None

    def header_str(self, x: str | bytes):
        res = self.header(x)
        if res is None:
            return None

        return res.decode("latin-1")

    def add_request_span_attrs(self, data: AttributesDict, prefix: str):
        current_http_request_span().set_attributes(dict_to_attrs(data, prefix))

    @trace_app_function
    def add_db_response(self, data: AttributesDict):
        # todo(maximsmol): datadog has shit support for events
        # current_http_request_span().add_event(
        #     f"database response {self._db_response_idx}", dict_to_attrs(data, "data")
        # )
        self.add_request_span_attrs(data, f"db.response.{self._db_response_idx}")
        self._db_response_idx += 1

    @trace_app_function
    async def receive_request_payload(self, cls: type[T]) -> T:
        json, res = await receive_class_ext(self.receive, cls)

        # todo(maximsmol): datadog has shit support for events
        # current_http_request_span().add_event(
        #     "request payload", dict_to_attrs(json, "data")
        # )
        self.add_request_span_attrs(json, "http.request_payload")

        return res


HandlerResult: TypeAlias = Any | None
Handler: TypeAlias = Callable[
    [Context],
    Awaitable[HandlerResult],
]
Route: TypeAlias = Handler | tuple[list[HTTPMethod], Handler]



import dataclasses
from dataclasses import dataclass
from typing import (
    Awaitable,
    Callable,
    Iterable,
    Literal,
    TypeAlias,
    get_args,
    get_origin,
)

import hypercorn.typing as htyping
from latch_data_validation.data_validation import validate


def type_str(x: type) -> str:
    for f in dataclasses.fields(x):
        if f.name != "type":
            continue

        o = get_origin(f.type)
        if o is not Literal:
            raise ValueError("'type' field type is not a Literal")

        res = get_args(f.type)[0]
        if not isinstance(res, str):
            raise ValueError("'type' field Literal is not a string")

        return res

    raise ValueError("'type' field not found")


@dataclass(frozen=True)
class ASGIVersions:
    spec_version: str
    version: Literal["2.0"] | Literal["3.0"]

    def as_dict(self):
        return validate(dataclasses.asdict(self), htyping.ASGIVersions)


# >>> Lifespan
@dataclass(frozen=True)
class LifespanScope:
    type: Literal["lifespan"]
    asgi: ASGIVersions

    def as_dict(self):
        return validate(dataclasses.asdict(self), htyping.LifespanScope)


@dataclass(frozen=True)
class LifespanStartupEvent:
    type: Literal["lifespan.startup"]

    def as_dict(self):
        return validate(dataclasses.asdict(self), htyping.LifespanStartupEvent)


@dataclass(frozen=True)
class LifespanShutdownEvent:
    type: Literal["lifespan.shutdown"]

    def as_dict(self):
        return validate(dataclasses.asdict(self), htyping.LifespanShutdownEvent)


LifespanReceiveEvent: TypeAlias = LifespanStartupEvent | LifespanShutdownEvent
LifespanReceiveCallable: TypeAlias = Callable[[], Awaitable[LifespanReceiveEvent]]


@dataclass(frozen=True)
class LifespanStartupCompleteEvent:
    type: Literal["lifespan.startup.complete"]

    def as_dict(self):
        return validate(dataclasses.asdict(self), htyping.LifespanStartupCompleteEvent)


@dataclass(frozen=True)
class LifespanStartupFailedEvent:
    type: Literal["lifespan.startup.failed"]
    message: str

    def as_dict(self):
        return validate(dataclasses.asdict(self), htyping.LifespanStartupFailedEvent)


LifespanStartupSendEvent: TypeAlias = (
    LifespanStartupCompleteEvent | LifespanStartupFailedEvent
)


@dataclass(frozen=True)
class LifespanShutdownCompleteEvent:
    type: Literal["lifespan.shutdown.complete"]

    def as_dict(self):
        return validate(dataclasses.asdict(self), htyping.LifespanShutdownCompleteEvent)


@dataclass(frozen=True)
class LifespanShutdownFailedEvent:
    type: Literal["lifespan.shutdown.failed"]
    message: str

    def as_dict(self):
        return validate(dataclasses.asdict(self), htyping.LifespanShutdownFailedEvent)


LifespanShutdownSendEvent: TypeAlias = (
    LifespanShutdownCompleteEvent | LifespanShutdownFailedEvent
)

LifespanSendEvent: TypeAlias = LifespanStartupSendEvent | LifespanShutdownSendEvent
LifespanSendCallable: TypeAlias = Callable[[LifespanSendEvent], Awaitable[None]]


# >>> HTTP
@dataclass(frozen=True)
class HTTPScope:
    type: Literal["http"]
    asgi: ASGIVersions
    http_version: str
    method: str
    scheme: str
    path: str
    raw_path: bytes
    query_string: bytes
    root_path: str
    headers: Iterable[tuple[bytes, bytes]]
    client: tuple[str, int] | None
    server: tuple[str, int | None] | None
    extensions: dict[str, dict]

    def as_dict(self):
        return validate(dataclasses.asdict(self), htyping.HTTPScope)


@dataclass(frozen=True)
class HTTPRequestEvent:
    type: Literal["http.request"]
    body: bytes
    more_body: bool

    def as_dict(self):
        return validate(dataclasses.asdict(self), htyping.HTTPRequestEvent)


@dataclass(frozen=True)
class HTTPDisconnectEvent:
    type: Literal["http.disconnect"]

    def as_dict(self):
        return validate(dataclasses.asdict(self), htyping.HTTPDisconnectEvent)


HTTPReceiveEvent: TypeAlias = HTTPRequestEvent | HTTPDisconnectEvent
HTTPReceiveCallable: TypeAlias = Callable[[], Awaitable[HTTPReceiveEvent]]


@dataclass(frozen=True)
class HTTPResponseStartEvent:
    type: Literal["http.response.start"]
    status: int
    headers: Iterable[tuple[bytes, bytes]]

    def as_dict(self):
        return validate(dataclasses.asdict(self), htyping.HTTPResponseStartEvent)


@dataclass(frozen=True)
class HTTPResponseBodyEvent:
    type: Literal["http.response.body"]
    body: bytes
    more_body: bool

    def as_dict(self):
        return validate(dataclasses.asdict(self), htyping.HTTPResponseBodyEvent)


@dataclass(frozen=True)
class HTTPServerPushEvent:
    type: Literal["http.response.push"]
    path: str
    headers: Iterable[tuple[bytes, bytes]]

    def as_dict(self):
        return validate(dataclasses.asdict(self), htyping.HTTPServerPushEvent)


HTTPSendEvent: TypeAlias = (
    HTTPResponseStartEvent
    | HTTPResponseBodyEvent
    | HTTPServerPushEvent
    | HTTPDisconnectEvent
)
HTTPSendCallable: TypeAlias = Callable[[HTTPSendEvent], Awaitable[None]]



# todo(maximsmol): move JWT stuff into envoy

import re
from dataclasses import dataclass
from http import HTTPStatus
from typing import Literal

import jwt
from jwt import PyJWKClient
from latch_o11y.o11y import app_tracer, trace_app_function

from .config import config
from .framework import HTTPErrorResponse

jwk_client = PyJWKClient("https://latchai.us.auth0.com/.well-known/jwks.json")
debug_authentication_header_regex = re.compile(
    r"""
    ^(
        Latch-Spoof-Sub \s+ (?P<oauth_sub>.*)
    )$
    """,
    re.IGNORECASE | re.VERBOSE,
)
authentication_header_regex = re.compile(
    r"""
    ^(
        Bearer \s+ (?P<oauth_token>.*) |
        Latch-Execution-Token \s+ (?P<execution_token>.*) |
        Latch-SDK-Token \s+ (?P<sdk_token>.*)
    )$
    """,
    re.IGNORECASE | re.VERBOSE,
)


class _HTTPUnauthorized(HTTPErrorResponse):
    """WARNING: HTTPForbidden is the correct error to use in virtually all cases"""

    def __init__(
        self,
        error_description: str,
        error: Literal["invalid_request"]
        | Literal["invalid_token"]
        | Literal["insufficient_scope"],
    ):
        escaped_description = error_description.replace('"', '\\"')
        super().__init__(
            HTTPStatus.UNAUTHORIZED,
            error_description,
            headers={
                "WWW-Authenticate": (
                    f'Bearer error="{error}", error_description="{escaped_description}"'
                )
            },
        )


@dataclass(frozen=True, kw_only=True)
class Authorization:
    oauth_sub: str | None = None
    execution_token: str | None = None
    sdk_token: str | None = None

    def unauthorized_if_none(self):
        if self.oauth_sub is not None:
            return
        if self.execution_token is not None:
            return
        if self.sdk_token is not None:
            return

        raise _HTTPUnauthorized(
            "Authenticaton required",
            error="invalid_request",
        )


@trace_app_function
def get_signer_sub(auth_header: str) -> Authorization:
    if auth_header is None:
        return Authorization()

    # todo(maximsmol): allow spoofing in prod if using a secret key
    # todo(maximsmol): allow spoofing using account ID instead of auth0 sub
    if config.allow_spoofing:
        auth_match = debug_authentication_header_regex.match(auth_header)
        if auth_match is not None:
            return Authorization(oauth_sub=auth_match.group("oauth_sub"))

    with app_tracer.start_as_current_span("match regex"):
        auth_match = authentication_header_regex.match(auth_header)
        if auth_match is None:
            raise _HTTPUnauthorized(
                error_description=(
                    "The `Authorization` header did not match the expected format."
                    " Accepted schemes: `Bearer` and `Latch-Execution-Token`"
                ),
                error="invalid_token",
            )

        execution_token = auth_match.group("execution_token")
        if execution_token is not None:
            return Authorization(execution_token=execution_token)

        sdk_token = auth_match.group("sdk_token")
        if sdk_token is not None:
            return Authorization(sdk_token=sdk_token)

        oauth_token = auth_match.group("oauth_token")

        if oauth_token is None:
            raise _HTTPUnauthorized(
                error_description="Could not parse the OAuth token",
                error="invalid_token",
            )

    with app_tracer.start_as_current_span("fetch jwk"):
        try:
            jwt_key = jwk_client.get_signing_key_from_jwt(oauth_token).key
        except jwt.exceptions.InvalidTokenError as e:
            raise _HTTPUnauthorized(
                error_description="JWT decoding failed",
                error="invalid_token",
            ) from e
        except jwt.exceptions.PyJWKClientError:
            # fixme(maximsmol): gut this abomination
            jwt_key = config.self_signed_jwk
            # raise _HTTPUnauthorized(
            #     error_description="No matching JWK found",
            #     error="invalid_token",
            # ) from e

    with app_tracer.start_as_current_span("decode jwt"):
        try:
            jwt_data: dict[str, str] = jwt.decode(
                oauth_token,
                key=jwt_key,
                algorithms=["RS256", "HS256"],
                # fixme(maximsmol): gut this abomination
                audience=(
                    config.audience
                    if jwt_key != config.self_signed_jwk
                    else None
                ),
            )
        except jwt.exceptions.InvalidTokenError as e:
            # todo(maximsmol): filter out scope failures and include the correct error code
            raise _HTTPUnauthorized(
                error_description="The JWT failed signature verification",
                error="invalid_token",
            ) from e

    return Authorization(oauth_sub=jwt_data["sub"])



from dataclasses import dataclass
from latch_config.config import read_config


@dataclass(frozen=True)
class AuthConfig:
    audience: str
    self_signed_jwk: str
    allow_spoofing: bool = False

config = read_config(AuthConfig, "auth")



import asyncio
import dataclasses
import traceback
from http import HTTPStatus

import opentelemetry.context as context
from hypercorn.typing import ASGIReceiveCallable, ASGISendCallable, Scope
from typing import Awaitable
from latch_data_validation.data_validation import untraced_validate, validate
from latch_o11y.o11y import log
from opentelemetry.propagate import set_global_textmap
from opentelemetry.trace import get_current_span, get_tracer

from .datadog_propagator import DDTraceContextTextMapPropagator

from .asgi_iface import (
    HTTPDisconnectEvent,
    HTTPReceiveCallable,
    HTTPRequestEvent,
    HTTPScope,
    HTTPSendCallable,
    HTTPSendEvent,
    LifespanReceiveCallable,
    LifespanScope,
    LifespanSendCallable,
    LifespanSendEvent,
    LifespanShutdownCompleteEvent,
    LifespanShutdownEvent,
    LifespanShutdownFailedEvent,
    LifespanStartupCompleteEvent,
    LifespanStartupEvent,
    LifespanStartupFailedEvent,
    type_str,
)
from .context import Context, Route
from .framework import (
    HTTPErrorResponse,
    HTTPInternalServerError,
    current_http_request_span,
    http_request_span_key,
    send_auto,
    send_data,
)

tracer = get_tracer(__name__)


# todo(maximsmol): ASGI instrumentation should trace lifespan

class LatchASGIServer:
    routes: dict[str, Route]
    startup_tasks: list[Awaitable] = []
    shutdown_tasks: list[Awaitable] = []

    def __init__(self, routes: dict[str, Route], startup_tasks: list[Awaitable] = [], shutdown_tasks: list[Awaitable] = []):
        self.routes = routes
        self.startup_tasks = startup_tasks
        self.shutdown_tasks = shutdown_tasks

    async def scope_lifespan(
        self, scope: LifespanScope, receive: LifespanReceiveCallable, send: LifespanSendCallable
    ):
        await log.info(
            f"Waiting for lifespan events (ASGI v{scope.asgi.version} @ spec"
            f" v{scope.asgi.spec_version})"
        )
        while True:
            message = await receive()
            await log.info(repr(message.type))

            if isinstance(message, LifespanStartupEvent):
                with tracer.start_as_current_span("startup"):
                    try:
                        await log.info("Connecting to Vacuole")
                        # todo(maximsmol): debug clock skew on connection reset
                        await asyncio.gather(*self.startup_tasks)

                        with tracer.start_as_current_span("send completion event"):
                            await send(
                                LifespanStartupCompleteEvent("lifespan.startup.complete")
                            )
                    except Exception as e:
                        with tracer.start_as_current_span("send failure event"):
                            await send(
                                LifespanStartupFailedEvent(
                                    "lifespan.startup.failed", str(e)
                                )
                            )

                        raise e
            elif isinstance(message, LifespanShutdownEvent):
                with tracer.start_as_current_span("shutdown"):
                    try:
                        await asyncio.gather(*self.shutdown_tasks)

                        with tracer.start_as_current_span("send completion event"):
                            await send(
                                LifespanShutdownCompleteEvent("lifespan.shutdown.complete")
                            )
                    except Exception as e:
                        with tracer.start_as_current_span("send failure event"):
                            await send(
                                LifespanShutdownFailedEvent(
                                    "lifespan.shutdown.failed", str(e)
                                )
                            )

                        raise e

                    break


    async def scope_http(
        self, scope: HTTPScope, receive: HTTPReceiveCallable, send: HTTPSendCallable
    ):
        ctx_reset_token: object | None = None
        try:
            new_ctx = context.set_value(http_request_span_key, get_current_span())
            ctx_reset_token = context.attach(new_ctx)

            current_http_request_span().set_attribute("resource.name", scope.path)
            await log.info(f"{scope.method} {scope.path}")

            with tracer.start_as_current_span("find route handler"):
                route = self.routes.get(scope.path)

            if not isinstance(route, tuple):
                methods = ["POST"]
                handler = route
            else:
                methods, handler = route

            if scope.method not in methods:
                if len(methods) == 1:
                    methods_str = methods[0]
                elif len(methods) == 2:
                    methods_str = f"{methods[0]} and {methods[1]}"
                else:
                    methods_str = ", and ".join([", ".join(methods[:-1]), methods[-1]])

                await send_data(
                    send,
                    HTTPStatus.METHOD_NOT_ALLOWED,
                    f"Only {methods_str} requests are supported",
                )
                return

            if handler is None:
                # todo(maximsmol): better error message
                await log.info("Not found")
                await send_data(send, HTTPStatus.NOT_FOUND, "Not found")
                return

            try:
                try:
                    ctx = Context(scope, receive, send)
                    res = await handler(ctx)

                    if res is not None:
                        with tracer.start_as_current_span("send response"):
                            await send_auto(send, HTTPStatus.OK, res)
                    return
                except HTTPErrorResponse as e:
                    raise e
                except Exception as e:
                    # todo(maximsmol): better error message
                    raise HTTPInternalServerError("Internal error") from e
            except HTTPErrorResponse as e:
                await send_auto(send, e.status, {"error": e.data}, headers=e.headers)
                if e.status == HTTPStatus.INTERNAL_SERVER_ERROR:
                    # await log.exception() # fixme(maximsmol)
                    traceback.print_exc()
                return
        finally:
            if ctx_reset_token is not None:
                context.detach(ctx_reset_token)


    async def raw_app(self, scope: Scope, receive: ASGIReceiveCallable, send: ASGISendCallable):
        try:
            if scope["type"] == "lifespan":

                async def ls_receive():
                    x = await receive()

                    if x["type"] == type_str(LifespanStartupEvent):
                        return untraced_validate(x, LifespanStartupEvent)

                    if x["type"] == type_str(LifespanShutdownEvent):
                        return untraced_validate(x, LifespanShutdownEvent)

                    raise RuntimeError(f"unknown lifespan event type: {repr(x['type'])}")

                async def ls_send(e: LifespanSendEvent):
                    data = dataclasses.asdict(e)
                    await send(data)

                return await self.scope_lifespan(
                    untraced_validate(scope, LifespanScope), ls_receive, ls_send
                )

            if scope["type"] == "http":

                async def http_receive():
                    x = await receive()

                    if x["type"] == type_str(HTTPRequestEvent):
                        return validate(x, HTTPRequestEvent)

                    if x["type"] == type_str(HTTPDisconnectEvent):
                        return validate(x, HTTPDisconnectEvent)

                    raise RuntimeError(f"unknown http event type: {repr(x['type'])}")

                async def http_send(e: HTTPSendEvent):
                    data = dataclasses.asdict(e)
                    await send(data)

                return await self.scope_http(validate(scope, HTTPScope), http_receive, http_send)

            raise RuntimeError(f"unsupported protocol: {repr(scope['type'])}")
        except Exception as e:
            await log.exception("Fallback exception handler:")
            raise e


set_global_textmap(DDTraceContextTextMapPropagator())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['latch_asgi']

package_data = \
{'': ['*']}

install_requires = \
['PyJWT[crypto]>=2.6.0,<3.0.0',
 'hypercorn[uvloop]>=0.14.3,<0.15.0',
 'latch-config>=0.1.6,<0.2.0',
 'latch-data-validation>=0.1.3,<0.2.0',
 'latch-o11y>=0.1.4,<0.2.0',
 'opentelemetry-api>=1.15.0,<2.0.0',
 'opentelemetry-instrumentation-asgi>=0.36b0,<0.37',
 'opentelemetry-sdk>=1.15.0,<2.0.0',
 'orjson>=3.8.5,<4.0.0',
 'pysimdjson>=5.0.2,<6.0.0']

setup_kwargs = {
    'name': 'latch-asgi',
    'version': '0.1.1',
    'description': 'ASGI python server',
    'long_description': '# python-asgi\n',
    'author': 'Max Smolin',
    'author_email': 'max@latch.bio',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)
